import numpy as np
from einops import rearrange
from flax import nnx
from jax import numpy as jnp


def _timestep_embedding(timesteps, embedding_dim: int, dtype=jnp.float32):
  half = embedding_dim // 2
  freqs = jnp.exp(-jnp.log(10_000) * jnp.arange(0, half) / half)
  emb = timesteps.astype(dtype)[:, None] * freqs[None, ...]
  emb = jnp.concatenate([jnp.sin(emb), jnp.cos(emb)], axis=1)
  return emb


def _modulate(inputs, shift, scale):
  return inputs * (1.0 + scale[:, None]) + shift[:, None]


class MLP(nnx.Module):
  def __init__(self, in_features, output_features, dropout_rate=None, *, rngs):
    features = [in_features] + output_features
    layers = []
    for index, (din, dout) in enumerate(zip(features[:-1], features[1:])):
      layers.append(nnx.Linear(in_features=din, out_features=dout, rngs=rngs))
    self.layers = tuple(layers)
    self.dropout_rate = dropout_rate
    if dropout_rate is not None:
      self.dropout_layer = nnx.Dropout(dropout_rate, rngs=rngs)

  def __call__(self, inputs):
    num_layers = len(self.layers)
    out = inputs
    for i, layer in enumerate(self.layers):
      out = layer(out)
      if i < num_layers - 1:
        out = nnx.gelu(out)
        if self.dropout_rate is not None:
          out = self.dropout_layer(out)
    return out


class OutProjection(nnx.Module):
  def __init__(self, hidden_size, patch_size, out_channels, *, rngs):
    super().__init__()
    self.norm = nnx.LayerNorm(hidden_size, rngs=rngs)
    self.out = nnx.Linear(
      hidden_size, patch_size * patch_size * out_channels, rngs=rngs
    )
    self.ada = nnx.Sequential(
      nnx.swish, nnx.Linear(hidden_size, 2 * hidden_size, rngs=rngs)
    )

  def __call__(self, inputs, context):
    shift, scale = jnp.split(self.ada(context), 2, -1)
    outs = self.out(_modulate(self.norm(inputs), shift, scale))
    return outs


class DiTBlock(nnx.Module):
  def __init__(
    self,
    hidden_size: int,
    n_heads: int,
    dropout_rate: float = 0.1,
    *,
    rngs: nnx.rnglib.Rngs,
  ):
    super().__init__()
    self.ada = nnx.Linear(hidden_size, hidden_size * 6, rngs=rngs)
    self.layer_norm1 = nnx.LayerNorm(
      hidden_size, use_scale=False, use_bias=False, rngs=rngs
    )
    self.self_attn = nnx.MultiHeadAttention(
      num_heads=n_heads, in_features=hidden_size, rngs=rngs
    )
    self.layer_norm2 = nnx.LayerNorm(
      hidden_size, use_scale=False, use_bias=False, rngs=rngs
    )
    self.mlp = MLP(
      hidden_size, [hidden_size * 4, hidden_size], dropout_rate, rngs=rngs
    )

  def __call__(self, inputs, context, **kwargs):
    hidden = inputs
    adaln_norm = self.ada(context)
    attn, gate = jnp.split(adaln_norm, 2, axis=-1)

    pre_shift, pre_scale, post_scale = jnp.split(attn, 3, -1)
    intermediate = _modulate(self.layer_norm1(hidden), pre_shift, pre_scale)
    intermediate = self.self_attn(intermediate)
    hidden = hidden + post_scale[:, None] * intermediate

    pre_shift, pre_scale, post_scale = jnp.split(gate, 3, -1)
    intermediate = _modulate(self.layer_norm2(hidden), pre_shift, pre_scale)
    intermediate = self.mlp(intermediate)
    outputs = hidden + post_scale[:, None] * intermediate

    return outputs


class DiT(nnx.Module):
  n_channels: int
  n_out_channels: int
  patch_size: int
  n_blocks: int
  n_heads: int
  dropout_rate: float = 0.1

  def __init__(
    self,
    n_in_channels,
    n_channels,
    n_out_channels,
    patch_size,
    n_blocks,
    n_heads,
    dropout_rate,
    *,
    rngs,
  ):
    self.patch_size = patch_size
    self.time_embedding = nnx.Sequential(
      nnx.Linear(n_channels, n_channels, rngs=rngs),
      nnx.swish,
      nnx.Linear(n_channels, n_channels, rngs=rngs),
      nnx.swish,
    )
    self.patchify = nnx.Conv(
      n_in_channels,
      n_channels,
      (patch_size, patch_size),
      (patch_size, patch_size),
      padding="VALID",
      kernel_init=nnx.initializers.xavier_uniform(),
      rngs=rngs,
    )
    self.patch_embedding = nnx.Param(
      nnx.initializers.lecun_normal()(
        rngs.params(), (1, n_channels, n_channels)
      )
    )
    self.dit_blocks = tuple(
      [DiTBlock(n_channels, n_heads, dropout_rate, rngs=rngs) for _ in n_blocks]
    )
    self.final_time_embedding = nnx.Linear(
      n_channels, n_channels * 2, rngs=rngs
    )
    self.out_projection = OutProjection(
      n_channels, patch_size, n_out_channels, rngs=rngs
    )

  def _patchify(self, inputs):
    B, H, W, C = inputs.shape
    n_patches = H // self.patch_size
    hidden = self.patchify(inputs)
    outputs = rearrange(
      hidden, "b h w c -> b (h w) c", h=n_patches, w=n_patches
    )
    return outputs

  def _unpatchify(self, inputs):
    B, HW, *_ = inputs.shape
    h = w = int(np.sqrt(HW))
    p = q = self.patch_size
    hidden = jnp.reshape(inputs, (B, h, w, p, q, self.n_out_channels))
    outputs = rearrange(
      hidden, "b h w p q c -> b (h p) (w q) c", h=h, w=w, p=q, q=q
    )
    return outputs

  def _embed(self, inputs):
    return inputs + self.patch_embedding.value

  def __call__(self, inputs, times):
    hidden = self._patchify(inputs)
    hidden = self._embed(hidden)
    times = self.time_embedding(_timestep_embedding(times, self.n_channels * 2))

    for block in self.dit_blocks:
      hidden = block(hidden, context=times)

    hidden = self.out_projection(hidden, times)
    outputs = self._unpatchify(hidden)
    return outputs


def SmallDiT(**kwargs):
  return DiT(n_channels=384, patch_size=2, n_blocks=12, n_heads=6, **kwargs)



from functools import partial

import optax
from flax import nnx


class CNN(nnx.Module):
  """Copy==pasted from the Flax documentation."""

  def __init__(self, *, rngs: nnx.Rngs):
    self.conv1 = nnx.Conv(1, 32, kernel_size=(3, 3), rngs=rngs)
    self.batch_norm1 = nnx.BatchNorm(32, rngs=rngs)
    self.dropout1 = nnx.Dropout(rate=0.025, rngs=rngs)
    self.conv2 = nnx.Conv(32, 64, kernel_size=(3, 3), rngs=rngs)
    self.batch_norm2 = nnx.BatchNorm(64, rngs=rngs)
    self.avg_pool = partial(nnx.avg_pool, window_shape=(2, 2), strides=(2, 2))
    self.linear1 = nnx.Linear(3136, 256, rngs=rngs)
    self.dropout2 = nnx.Dropout(rate=0.025, rngs=rngs)
    self.linear2 = nnx.Linear(256, 10, rngs=rngs)

  def __call__(self, x):
    x = self.avg_pool(nnx.relu(self.batch_norm1(self.dropout1(self.conv1(x)))))
    x = self.avg_pool(nnx.relu(self.batch_norm2(self.conv2(x))))
    x = x.reshape(x.shape[0], -1)
    x = nnx.relu(self.dropout2(self.linear1(x)))
    x = self.linear2(x)
    return x


def loss_fn(model, images, labels):
  logits = model(images)
  return optax.losses.softmax_cross_entropy_with_integer_labels(
    logits=logits, labels=labels
  ).mean()


def train_step(model, rng_key, batch, **kwargs):
  return nnx.value_and_grad(loss_fn)(model, batch["image"], batch["label"])


def val_step(model, rng_key, batch, **kwargs):
  return loss_fn(model, batch["image"], batch["label"])



import os

import jax
import orbax.checkpoint as ocp
from absl import logging
from flax import nnx


def get_default_checkpointer(
  outfolder,
  *,
  save_frequency=1,
  max_to_keep=5,
  best_mode="min",
  best_fn=lambda x: x["val/loss"],
):
  checkpointer = ocp.PyTreeCheckpointer()
  options = ocp.CheckpointManagerOptions(
    max_to_keep=max_to_keep, create=True, best_mode=best_mode, best_fn=best_fn
  )
  checkpoint_manager = ocp.CheckpointManager(
    os.path.join(outfolder, "best"),
    options=options,
    item_names=("state", "opt_state"),
  )

  def save_fn(step, *, model, optimizer, metrics):
    if step % save_frequency != 0:
      return

    _, state = nnx.split(model)
    _, opt_state = nnx.split(optimizer.opt_state)
    try:
      logging.info("saving best checkpoint")
      checkpoint_manager.save(
        step=step,
        args=ocp.args.Composite(
          state=ocp.args.StandardSave(state),
          opt_state=ocp.args.StandardSave(opt_state),
        ),
        metrics=metrics,
        force=True,
      )
      checkpoint_manager.wait_until_finished()
    except Exception as e:
      logging.error(f"could not save checkpoint because of: {e}")
      logging.error("resuming nonetheless")
    try:
      logging.info("saving last checkpoint")
      checkpointer.save(
        os.path.join(outfolder, "last"), (state, opt_state), force=True
      )
    except Exception as e:
      logging.error(f"could not last checkpoint because of: {e}")
      logging.error("resuming nonetheless")

  def restore_best_fn(model, optimizer):
    graph_def, state = nnx.split(model)
    opt_def, opt_state = nnx.split(optimizer.opt_state)
    restored = checkpoint_manager.restore(
      checkpoint_manager.best_step(),
      args=ocp.args.Composite(
        state=ocp.args.StandardRestore(nnx.eval_shape(lambda: state)),
        opt_state=ocp.args.StandardRestore(nnx.eval_shape(lambda: opt_state)),
      ),
    )
    model = nnx.merge(graph_def, restored["state"])
    optimizer.opt_state = nnx.merge(opt_def, restored["opt_state"])
    return model, optimizer

  def restore_last_fn(model, optimizer):
    graphdef, state = nnx.split(model)
    optdef, opt_state = nnx.split(optimizer.opt_state)
    state = nnx.eval_shape(lambda: state)
    opt_state = nnx.eval_shape(lambda: opt_state)
    try:
      logging.info("trying to restore last checkpoint")
      restored = checkpointer.restore(
        os.path.join(outfolder, "last"), (state, opt_state)
      )
      model = nnx.merge(graphdef, restored[0])
      optimizer.opt_state = nnx.merge(optdef, restored[1])
    except FileNotFoundError:
      logging.warning("could not find last checkpoint. resuming with blank state")
    return model, optimizer

  return save_fn, restore_best_fn, restore_last_fn



import grain.python as grain
import tensorflow_datasets as tfds
from grain._src.core.transforms import Batch
from jax import numpy as jnp
from jax import random as jr


def data_loaders(
  rng_key,
  outfolder,
  *,
  n_epochs=100,
  n_workers=0,
  batch_size=128,
  buffer_size=1,
  shuffle=True,
  split="train",
):
  datasets = tfds.data_source(
    "mnist",
    split=split,
    data_dir=outfolder,
    download_and_prepare_kwargs={"file_format": "parquet"},
  )
  if isinstance(split, str):
    datasets = [datasets]
  itrs = []
  for dataset in datasets:
    itr_key, rng_key = jr.split(rng_key)
    itr = _as_grain_data_loader(
      itr_key, dataset, n_epochs, n_workers, batch_size, buffer_size, shuffle
    )
    itrs.append(itr)
  return itrs


def _as_grain_data_loader(
  rng_key, itr, n_epochs, n_workers, batch_size, buffer_size, shuffle
):
  max_int32 = jnp.iinfo(jnp.int32).max
  seed = jr.randint(rng_key, shape=(), minval=0, maxval=max_int32)

  index_sampler = grain.IndexSampler(
    num_records=len(itr),
    num_epochs=n_epochs,
    shard_options=grain.ShardOptions(
      shard_index=0, shard_count=1, drop_remainder=True
    ),
    shuffle=shuffle,
    seed=int(seed),
  )
  data_loader = grain.DataLoader(
    data_source=itr,
    operations=[Batch(batch_size, drop_remainder=True)],
    shard_options=grain.ShardOptions(
      shard_index=0, shard_count=1, drop_remainder=True
    ),
    sampler=index_sampler,
    worker_count=n_workers,
  )
  return data_loader



"""fll: A high-level API for building and training Flax NNX models."""

__version__ = "0.0.1"

from blaxbird._src.checkpointer import get_default_checkpointer
from blaxbird._src.trainer import train_fn

__all__ = ["get_default_checkpointer", "train_fn"]



import os

import dataloader
import jax
import optax
from flax import nnx
from jax import random as jr
from jax.experimental import mesh_utils
from train_state import get_optimizer

from examples.flow_matching import FlowMatching
from examples.model import SmallDiT
from fll import get_default_checkpointer, train_fn


def get_optimizer(model, lr=1e-04):
  tx = optax.adamw(lr)
  tx = nnx.Optimizer(model, tx=tx)
  return tx


def get_sharding():
  num_devices = jax.local_device_count()
  mesh = jax.sharding.Mesh(
    mesh_utils.create_device_mesh((num_devices,)), ("data",)
  )
  model_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec())
  data_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec("data"))
  return model_sharding, data_sharding


def get_hooks(module, outfolder):
  checkpointer = get_default_checkpointer(outfolder, module)
  return [checkpointer]


def get_train_and_val_iters(rng_key, outfolder):
  return dataloader.data_loaders(rng_key, outfolder)


def run():
  outfolder = "./"
  rngs = nnx.rnglib.Rngs(jr.key(0))

  model = SmallDiT(rngs=rngs)
  optimizer = get_optimizer(model)
  model_sharding, data_sharding = get_sharding()
  module = FlowMatching(model, optimizer=optimizer, sharding=model_sharding)

  hooks = [
    get_default_checkpointer(module, os.path.join(outfolder, "checkpoints"))
  ]
  train_itr, val_itr = get_train_and_val_iters(
    jr.key(1), os.path.join(outfolder, "data")
  )

  train = train_fn(
    n_steps=100,
    n_eval_frequency=10,
    n_eval_batches=10,
    sharding=data_sharding,
    hooks=hooks,
    log_to_wandb=True,
  )
  train(jax.key(2), module, train_itr, val_itr)


if __name__ == "__main__":
  run()



import dataclasses

import numpy as np
from flax import nnx
from jax import numpy as jnp
from jax import random as jr

import fll


def _forward_process(inputs, times, noise):
  new_shape = (-1,) + tuple(np.ones(inputs.ndim - 1, dtype=np.int32).tolist())
  times = times.reshape(new_shape)
  inputs_t = times * inputs + (1.0 - times) * noise
  return inputs_t


@dataclasses.dataclass
class FlowMatchingConfig:
  n_sampling_steps: int
  time_eps: float = 1e-3
  time_max: float = 1.0


class FlowMatching(fll.Module):
  def __init__(
    self, model, config=FlowMatchingConfig(10), *, optimizer, sharding
  ):
    super().__init__(model, optimizer, sharding)
    self.config = config

  def __call__(self, inputs, times):
    return self.model(inputs=inputs, times=times * 999.0)

  def _loss_fn(self, rng_key, inputs):
    time_key, noise_key = jr.split(rng_key)
    times = jr.uniform(time_key, shape=(inputs.shape[0],))
    times = (
      times * (self.config.time_max - self.config.time_eps)
      + self.config.time_eps
    )
    noise = jr.normal(noise_key, inputs.shape)
    inputs_t = _forward_process(inputs, times, noise)
    vs = self(inputs_t, times)
    target = inputs - noise
    loss = jnp.mean(jnp.square(target - vs))
    return loss

  @nnx.jit
  def train_step(self, rng_key, batch):
    loss, grads = nnx.value_and_grad(self._loss_fn)((rng_key, batch["inputs"]))
    return loss, grads

  @nnx.jit
  def eval_step(self, rng_key, batch):
    loss = self._loss_fn(rng_key, batch["inputs"])
    return loss

  @nnx.jit
  def sample(self, rng_key, sample_shape):
    n = sample_shape[0]
    dt = 1.0 / self.config.n_sampling_steps
    samples = jr.normal(rng_key, sample_shape)
    for i in range(self.config.n_sampling_steps):
      times = i / self.config.n_sampling_steps
      times = (
        times * (self.config.time_max - self.config.time_eps)
        + self.config.time_eps
      )
      times = jnp.repeat(times, n)
      vt = self(
        inputs=samples,
        times=times * 999.0,
      )
      samples = samples + vt * dt
    return samples



from collections.abc import Callable, Iterable

import grain.python as grain
import jax
import optax
import wandb
from absl import logging
from flax import nnx
from jax import random as jr


def step_and_val_fns(fns):
  step, eval = fns

  def train_step(model, rng_key, optimizer, metrics, batch, **kwargs):
    model.train()
    loss, grads = step(model, rng_key, batch, **kwargs)
    optimizer.update(grads)
    metrics.update(loss=loss)
    return {"loss": loss}

  def eval_step(model, rng_key, metrics, batch, **kwargs):
    model.eval()
    loss = eval(model, rng_key, batch, **kwargs)
    metrics.update(loss=loss)
    return {"loss": loss}

  return train_step, eval_step


def train_fn(
  *,
  fns,
  shardings,
  n_steps,
  eval_frequency,
  n_eval_batches,
  log_to_wandb=False,
  hooks: Iterable[Callable] = (),
):
  def train(
    rng_key,
    model: nnx.Module,
    optimizer: optax.GradientTransformation,
    train_itr: grain.DataLoader,
    val_itr: grain.DataLoader,
  ):
    # get train and val fns
    step_fn, eval_fn = step_and_val_fns(fns)
    # get model and replicate
    state = nnx.state((model, optimizer))
    state = jax.device_put(state, shardings[0])
    nnx.update((model, optimizer), state)
    # metrics
    metrics = nnx.MultiMetric(loss=nnx.metrics.Average("loss"))
    metrics_history = {}
    # run training
    step_key, rng_key = jr.split(rng_key)
    for step, batch in zip(range(1, n_steps + 1), train_itr):
      train_key, val_key = jr.split(jr.fold_in(step_key, step))
      batch = jax.device_put(batch, shardings[1])
      # do a gradient step
      step_fn(
        model=model,
        rng_key=train_key,
        optimizer=optimizer,
        metrics=metrics,
        batch=batch,
      )
      is_first_step = step == 1
      is_last_step = step == n_steps
      is_first_or_last_step = is_first_step or is_last_step
      if step % eval_frequency == 0 or is_first_or_last_step:
        # store training losses
        for metric, value in metrics.compute().items():
          metrics_history[f"train/{metric}"] = float(value)
        # do evaluation loop
        for val_idx, batch in zip(range(n_eval_batches), val_itr):
          batch = jax.device_put(batch, shardings[1])
          eval_fn(
            model=model,
            rng_key=jr.fold_in(val_key, val_idx),
            metrics=metrics,
            batch=batch,
          )
        # store val losses
        for metric, value in metrics.compute().items():
          metrics_history[f"val/{metric}"] = float(value)
        metrics.reset()
        # log losses after each val round
        if jax.process_index() == 0:
          logging.info(
            f"loss at step {step}: "
            f"{metrics_history['train/loss']}/"
            f"{metrics_history['val/loss']}"
          )
        if log_to_wandb and jax.process_index() == 0:
          wandb.log(metrics_history, step=step)
      for h in hooks:
        h(step, model=model, optimizer=optimizer, metrics=metrics_history)

  return train



import os

import dataloader
import jax
import optax
from absl import logging
from flax import nnx
from jax import random as jr
from jax.experimental import mesh_utils
from blaxbird import train_fn, get_default_checkpointer

from model import CNN, train_step, val_step


def get_optimizer(model, lr=1e-4):
  tx = optax.adamw(lr)
  tx = nnx.Optimizer(model, tx=tx)
  return tx


def get_sharding():
  num_devices = jax.local_device_count()
  mesh = jax.sharding.Mesh(
    mesh_utils.create_device_mesh((num_devices,)), ("data",)
  )
  model_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec())
  data_sharding = jax.NamedSharding(mesh, jax.sharding.PartitionSpec("data"))
  return model_sharding, data_sharding


def visualize_hook(val_iter, n_eval_frequency):
  def hook_fn(metrics, val_iter, n_eval_frequency):
    def fn(step, *, model, **kwargs):
      if step % n_eval_frequency != 0:
        return
      batch = next(iter(val_iter))
      logits = model(batch["image"])
      loss = optax.softmax_cross_entropy_with_integer_labels(
        logits=logits, labels=batch["label"]
      ).mean()
      metrics.update(loss=loss, logits=logits, labels=batch["label"])
      if jax.process_index() == 0:
        curr_metrics = ", ".join(
          [f"{k}: {v}" for k, v in metrics.compute().items()]
        )
        logging.info(f"metrics at step {step}: {curr_metrics}")
        metrics.reset()

    return fn

  metrics = nnx.MultiMetric(
    accuracy=nnx.metrics.Accuracy(),
    loss=nnx.metrics.Average("loss"),
  )
  return hook_fn(metrics, val_iter, n_eval_frequency)


def get_hooks(val_itr, n_eval_frequency):
  return [visualize_hook(val_itr, n_eval_frequency)]


def get_train_and_val_itrs(rng_key, outfolder):
  return dataloader.data_loaders(
    rng_key, outfolder, split=["train[:90%]", "train[90%:]"]
  )


def run():
  logging.set_verbosity(logging.INFO)

  outfolder = os.path.dirname(__file__)
  n_steps, n_eval_frequency, n_eval_batches = 100, 10, 5
  train_itr, val_itr = get_train_and_val_itrs(
    jr.key(0), os.path.join(outfolder, "data")
  )

  model = CNN(rngs=nnx.rnglib.Rngs(jr.key(1)))
  optimizer = get_optimizer(model)
  model_sharding, data_sharding = get_sharding()
  hooks = get_hooks(val_itr, n_eval_frequency)

  save_fn, _, restore_last_fn = get_default_checkpointer(os.path.join(outfolder, "checkpoints"), save_frequency=2)
  model, optimizer = restore_last_fn(model, optimizer)
  hooks.append(save_fn)

  train = train_fn(
    fns=(train_step, val_step),
    n_steps=n_steps,
    eval_frequency=n_eval_frequency,
    n_eval_batches=n_eval_batches,
    shardings=(model_sharding, data_sharding),
    hooks=hooks,
    log_to_wandb=False,
  )
  train(jr.key(2), model, optimizer, train_itr, val_itr)


if __name__ == "__main__":
  run()
